from __future__ import annotations
import boto3
import io
import json
from abc import ABC, abstractmethod
from typing import Optional, Any, Dict, List
from datetime import datetime, date
from copy import copy
import os
import gzip
from enum import Enum
from my_aws_helpers.logging import select_powertools_logger


logger = select_powertools_logger("aws-helpers-s3")


class ContentType(str, Enum):
    plain_text = "text/plain"
    xml_content = "text/xml"
    json_content = "application/json"
    pdf_content = "application/pdf"
    jpeg_content = "image/jpeg"


class ContentEncoding(str, Enum):
    gzip = "gzip"


class S3Serialiser:

    @staticmethod
    def _serialise(obj: Any):
        if isinstance(obj, datetime) or isinstance(obj, date):
            return obj.isoformat()
        if isinstance(obj, S3Location):
            return obj.location
        return obj

    @staticmethod
    def object_serialiser(obj: Dict):
        if isinstance(obj, list):
            return [S3Serialiser.object_serialiser(obj=obj) for obj in obj]
        if isinstance(obj, dict):
            return {k: S3Serialiser.object_serialiser(v) for k, v in obj.items()}
        return S3Serialiser._serialise(obj=obj)


class S3Location:
    bucket: str
    file_name: str
    location: str

    @classmethod
    def from_location(cls, location: str):
        bucket, file_name = location.split("/")[0], "/".join(location.split("/")[1:])
        return cls(bucket=bucket, file_name=file_name)

    def __init__(self, bucket: str, file_name: str) -> None:
        self.bucket = bucket
        self.file_name = file_name
        self.location = f"{self.bucket}/{self.file_name}"

    def serialise(self):
        return copy(vars(self))


class S3:
    client: boto3.client

    def __init__(
        self,
        client: Optional[boto3.client] = None,
        resource: Optional[boto3.resource] = None,
    ) -> None:
        self.client = client if client else self._get_client()
        self.resource = (
            resource
            if resource
            else boto3.resource("s3", region_name=os.environ["AWS_DEFAULT_REGION"])
        )

    def _get_client(self) -> boto3.client:
        region_name = os.environ["AWS_DEFAULT_REGION"]
        s3_client = boto3.client("s3", region_name=region_name)
        endpoint_url = s3_client.meta.endpoint_url
        s3_client = boto3.client(
            "s3", region_name=region_name, endpoint_url=endpoint_url
        )
        return s3_client

    def _streaming_body_to_dict(self, payload):
        file_like_obj = io.BytesIO(payload.read())
        response = json.loads(file_like_obj.getvalue())
        return response

    def put_json_object(self, bucket_name: str, file_name: str, object: dict):
        return self.client.put_object(
            Body=json.dumps(object), Bucket=bucket_name, Key=file_name
        )

    def list_objects_by_prefix(self, bucket_name: str, prefix: str) -> List[S3Location]:
        try:
            response = self.client.list_objects_v2(Bucket=bucket_name, Prefix=prefix)
            contents = response["Contents"]
            locations = [
                S3Location(bucket=bucket_name, file_name=c["Key"]) for c in contents
            ]
            return locations
        except Exception as e:
            logger.exception(
                f"Failed to get objects from s3: {bucket_name}/{prefix} due to {e}"
            )
            return []

    def get_object(self, bucket_name: str, file_name: str):
        response = self.client.get_object(Bucket=bucket_name, Key=file_name)
        return self._streaming_body_to_dict(response["Body"])

    def get_presigned_url(
        self,
        bucket_name: str,
        file_name: str,
        expires_in: int = 3600,
    ):
        return self.client.generate_presigned_url(
            "get_object",
            Params={
                "Bucket": bucket_name,
                "Key": file_name,
            },
            ExpiresIn=expires_in,
        )

    def get_s3_location_from_bucket_file(
        bucket_name: str, file_name: str
    ) -> S3Location:
        return S3Location(bucket=bucket_name, file_name=file_name)

    def get_bucket_file_from_s3_location(s3_location: S3Location) -> S3Location:
        return S3Location.from_location(location=s3_location)

    def save_document_content(
        self,
        file_contents: bytes,
        s3_location: S3Location,
        content_encoding: str = "",
        content_type: str = "application/pdf",
        compress: bool = True,
    ) -> Optional[S3Location]:
        """
        saves document content to bucket, in file_name
        Options for content_type:
            "application/pdf"
            "text/plain"
            "application/json"
            probably more
        Options for content_encoding:
            "": default encoding
            "gzip": compressed contents
        """
        try:
            if compress or s3_location.file_name.endswith(".gz"):
                file_contents = gzip.compress(file_contents)
                content_encoding = ContentEncoding.gzip.value
            obj = self.resource.Object(s3_location.bucket, s3_location.file_name)
            obj.put(
                Body=file_contents,
                ContentType=content_type,
                ContentEncoding=content_encoding,
            )
            return s3_location
        except Exception as e:
            logger.exception(e)
            return None

    def read_binary_from_s3(self, s3_location: S3Location) -> Optional[bytes]:
        try:
            obj = self.resource.Object(s3_location.bucket, s3_location.file_name)
            d_bytes = io.BytesIO()
            obj.download_fileobj(d_bytes)
            d_bytes.seek(0)
            if obj.content_encoding == ContentEncoding.gzip.value:
                try:
                    with gzip.GzipFile(fileobj=d_bytes) as gz_file:
                        return gz_file.read()
                except gzip.BadGzipFile:
                    d_bytes.seek(0)
            return d_bytes.read()
        except Exception as e:
            logger.exception(f"Failed to read binary from s3 due to {e}")
            return None

    def save_text_to_s3(self, text: str, s3_location: S3Location):
        try:
            file_contents = bytes(text.encode("UTF-8"))
            return self.save_document_content(
                file_contents=file_contents,
                s3_location=s3_location,
                content_type=ContentType.plain_text.value,
                compress=True,
                content_encoding=ContentEncoding.gzip.value,
            )
        except Exception as e:
            logger.exception(f"Failed to save text to s3 due to {e}")
            return None

    def save_xml_to_s3(self, xml_text: bytes, s3_location: S3Location):
        """
        xml_text tends to come from:
            root = lxml.etree.ElementTree().get_root()
            xml_text = ET.tostring(root, encoding='utf-8')
        """
        try:
            return self.save_document_content(
                file_contents=xml_text,
                s3_location=s3_location,
                content_type=ContentType.xml_content.value,
                compress=True,
                content_encoding=ContentEncoding.gzip.value,
            )
        except Exception as e:
            logger.exception(f"Failed to save xml text to s3 due to {e}")
            return None

    def save_pdf_to_s3(self, pdf_content: bytes, s3_location: S3Location):
        """
        pdf_content tends to come from:
            PyMuPdf.Document().write()
        """
        try:
            return self.save_document_content(
                file_contents=pdf_content,
                s3_location=s3_location,
                content_type=ContentType.pdf_content.value,
                compress=True,
                content_encoding=ContentEncoding.gzip.value,
            )
        except Exception as e:
            logger.exception(f"Failed to save pdf to s3 due to {e}")
            return None

    def save_dict_to_s3(self, content: Dict[str, Any], s3_location: S3Location):
        """ """
        try:
            file_contents = bytes(json.dumps(content).encode("UTF-8"))
            return self.save_document_content(
                file_contents=file_contents,
                s3_location=s3_location,
                content_type=ContentType.json_content.value,
                compress=True,
                content_encoding=ContentEncoding.gzip.value,
            )
        except Exception as e:
            logger.exception(f"Failed to save dict to s3 due to {e}")
            return None

    def save_jpeg_to_s3(self, content: bytes, s3_location: S3Location):
        try:
            return self.save_document_content(
                file_contents=content,
                s3_location=s3_location,
                content_type=ContentType.jpeg_content.value,
                compress=True,
                content_encoding=ContentEncoding.gzip.value,
            )
        except Exception as e:
            logger.exception(f"Failed to save jpeg to s3 due to {e}")
            return None

    def read_dict_from_s3(self, s3_location: S3Location) -> dict:
        return json.loads(
            self.read_binary_from_s3(s3_location=s3_location).decode("utf-8")
        )


class BaseS3Object(ABC):
    def to_s3_representation(self) -> dict:
        obj = copy(vars(self))
        return S3Serialiser.object_serialiser(obj=obj)

    @classmethod
    def from_s3_representation(cls, obj: dict) -> BaseS3Object:
        return cls(**obj)

    @abstractmethod
    def get_save_location(self, bucket_name: str) -> S3Location:
        pass


class BaseS3Queries:
    s3_client: S3
    bucket_name: str

    def __init__(self, s3_client: S3, bucket_name: str):
        self.s3_client = s3_client
        self.bucket_name = bucket_name

    def save_s3_object_to_s3(self, object: BaseS3Object) -> Optional[S3Location]:
        try:
            obj = object.to_s3_representation()
            return self.s3_client.save_dict_to_s3(
                content=obj, s3_location=object.get_save_location()
            )
        except Exception as e:
            logger.exception(f"Failed to save s3 object to s3 due to {e}")
        return None



from __future__ import annotations
import boto3
from botocore.config import Config
import json
import time
import os
import io
import re
from copy import copy
from typing import Optional, List, Dict
from enum import Enum
import pymupdf
import concurrent.futures
from dataclasses import dataclass
from my_aws_helpers.s3 import S3Serialiser, BaseS3Object, BaseS3Queries, S3, S3Location

from my_aws_helpers.logging import select_powertools_logger


logger = select_powertools_logger("aws-helpers-bedrock")


class ImageType(str, Enum):
    gif = "gif"
    jpeg = "jpeg"
    png = "png"
    webp = "webp"


class PromptType(str, Enum):
    transaction_headers = "transactions_headers_prompt_v2.txt"
    transactions = "transactions_prompt.txt"
    # json = "json_system_prompt.txt"
    markdown = "markdown_system_prompt.txt"


@dataclass
class TokenUsage:
    input_tokens: int
    output_tokens: int
    total_tokens: int

    @classmethod
    def from_dict(cls, data: Dict[str, int]) -> TokenUsage:
        return cls(
            input_tokens=data.get("inputTokens", 0),
            output_tokens=data.get("outputTokens", 0),
            total_tokens=data.get("totalTokens", 0),
        )


@dataclass
class OCRResult(BaseS3Object):
    content: List[Dict[str, str]]
    token_usage: TokenUsage
    page_number: int

    @classmethod
    def from_dict(cls, data: Dict) -> OCRResult:
        return cls(
            content=data.get("content", []),
            token_usage=TokenUsage.from_dict(data.get("token_usage", {})),
            page_number=data.get("page_number", 0),
        )

    @classmethod
    def from_s3_representation(cls, obj: dict) -> OCRResult:
        obj['token_usage'] = TokenUsage.from_dict(obj.get("token_usage", {})),
        return cls(**obj)

    def to_s3_representation(self) -> dict:
        obj = copy(vars(self))
        obj['token_usage'] = S3Serialiser.object_serialiser(obj=vars(obj['token_usage']))
        return S3Serialiser.object_serialiser(obj=obj)

    def get_save_location(self, bucket_name: str) -> S3Location:
        pass


class OCRResultQueries(BaseS3Queries):
    def __init__(self, s3_client: S3, bucket_name: str):
        super().__init__(s3_client=s3_client, bucket_name=bucket_name)

    def save_ocr_result_to_s3(
        self, ocr_result: OCRResult, save_location: S3Location
    ) -> Optional[S3Location]:
        try:
            obj = ocr_result.to_s3_representation()
            return self.s3_client.save_dict_to_s3(
                content=obj,
                s3_location=save_location,
            )
        except Exception as e:
            logger.exception(f"Failed to save ocr result to s3 due to {e}")
        return None

    def _concurrent_s3_read(
        self, locations: List[S3Location], max_workers: int = 10
    ) -> List[OCRResult]:
        results: List[OCRResult] = list()
        futures = list()
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            for loc in locations:
                future = executor.submit(
                    self.s3_client.read_dict_from_s3,
                    s3_location=loc,
                )
                futures.append(future)
        for f in futures:
            results.append(f.result())
        results = [r for r in results if r is not None]
        return results

    def get_listing_by_key_prefix(self, prefix: str) -> List[OCRResult]:
        locations = self.s3_client.list_objects_by_prefix(
            bucket_name=self.bucket_name, prefix=prefix
        )
        objects = self._concurrent_s3_read(listing_locations=locations)
        ocr_results = [OCRResult.from_s3_representation(obj=obj) for obj in objects]
        return ocr_results


class Bedrock:
    def __init__(
        self,
        model_id: str = "apac.anthropic.claude-3-5-sonnet-20241022-v2:0",#anthropic.claude-sonnet-4-20250514-v1:0
        logger=None,
        sleep_time: float = 1.0,
    ):

        self.session = Bedrock._set_session_params()
        self.logger = logger
        region_name = "ap-southeast-2"
        if self.session is None:
            self.session = boto3.Session(region_name=region_name)
        self.sleep_time = sleep_time

        custom_config = Config(
            retries={
                "max_attempts": 2,  # Total attempts = 1 initial + 1 retry
                "mode": "standard",  # or 'adaptive'
            }
        )
        self.client = self.session.client(
            "bedrock-runtime", region_name=region_name, config=custom_config
        )
        self.model_id = model_id

    @staticmethod
    def _set_session_params():
        try:
            aws_access_key_id = os.environ["AWS_ACCESS_KEY_ID"]
            aws_secret_access_key = os.environ["AWS_SECRET_ACCESS_KEY"]
            aws_session_token = os.environ["AWS_SESSION_TOKEN"]
            region_name = os.environ["AWS_DEFAULT_REGION"]
            return boto3.Session(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                region_name=region_name,
            )
        except Exception as e:
            print(e)
            return None

    @staticmethod
    def extract_json_from_markdown(text: str):
        """
        Extracts the JSON object from a string that may be wrapped in ```json ... ``` code block
        """
        # Match a {...} block anywhere in the text
        match = re.search(r"\{.*\}", text, re.DOTALL)
        if match:
            json_str = match.group(0)
            return json.loads(json_str)
        else:
            raise ValueError("No JSON object found in the text")

    def _get_prompt(self, prompt_type: str) -> Optional[str]:
        if prompt_type not in list(PromptType):
            raise Exception(f"Error: Invalid prompt type")

        path = os.path.join(os.path.dirname(__file__), "prompts", prompt_type)
        try:
            with open(path, "r") as f:
                prompt = f.read()
                return prompt
        except Exception as e:
            self.logger.exception(f"Failed to get {prompt_type} prompt due to {e}")
        return None

    def _ocr(
        self, prompt: str, image_bytes: bytes, page_number: Optional[int] = 0
    ) -> Optional[OCRResult]:
        system_prompt = [{"text": prompt}]
        message = [
            {
                "role": "user",
                "content": [
                    {
                        "image": {
                            "format": "png",
                            "source": {
                                "bytes": image_bytes,
                            },
                        }
                    }
                ],
            }
        ]
        retries = 3
        for i in range(retries):
            self.logger.info(f"Attempt number {i} for {self.model_id} converse")
            try:
                response = self.client.converse(
                    modelId=self.model_id, messages=message, system=system_prompt
                )
                if response["ResponseMetadata"]["HTTPStatusCode"] == 200:
                    break
            except Exception as e:
                self.logger.exception(f"Error during conversation due to {e}")
                if i >= len(retries) - 1:
                    raise Exception(e)
                time.sleep(self.sleep_time)
                continue

        result = {}
        result["content"] = json.loads(
            response["output"]["message"]["content"][0]["text"]
        )
        result["token_usage"] = response["usage"]
        result["page_number"] = page_number
        return OCRResult.from_dict(data=result)

    def _parallel_ocr(
        self,
        image_bytes_list: List[bytes],
        prompt: str,
        max_workers: int = 10,
    ):
        execution_futures = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            for i, img in enumerate(image_bytes_list):
                self.logger.info(f"Starting OCR for page: {i}")
                time.sleep(self.sleep_time)  # Stagger start time
                future = executor.submit(self._ocr, prompt=prompt, image_bytes=img)
                execution_futures.append(future)

        # Wait for all tasks and collect results in order of submission
        results = [
            future.result()
            for future in execution_futures
            if future.result() is not None
        ]
        return results

    def get_ocr_result(
        self,
        pdf_bytes: io.BytesIO,
        prompt_type: str,
        zoom: int = 7,
    ) -> List[OCRResult]:
        try:
            self.logger.info("Getting OCR Results")
            document = pymupdf.open(stream=pdf_bytes, filetype="pdf")
            pages: List[pymupdf.Page] = [p for p in document]

            image_bytes_list: List[bytes] = list()
            for i, p in enumerate(pages):
                try:
                    image_bytes: bytes = p.get_pixmap(
                        matrix=pymupdf.Matrix(zoom, zoom)
                    ).tobytes("png")
                    image_bytes_list.append(image_bytes)
                except Exception as e:
                    self.logger.error(f"Could not get pix map for page {i}")
                    continue
            skip_page_zero = False
            header_ocr_result = None
            if len(image_bytes_list) > 1:
                headers_prompt = self._get_prompt(
                    prompt_type=PromptType.transaction_headers.value
                )
                for i in range(2):
                    # try to get headers from the first or second page
                    header_ocr_result = self._ocr(
                        prompt=headers_prompt, image_bytes=image_bytes_list[i]
                    )
                    if header_ocr_result is None:
                        self.logger.info(
                            f"No ocr result returned when getting headers {PromptType.transaction_headers.value}"
                        )
                    headers = header_ocr_result.content.get("headers")
                    if (len(headers) < 1) or (headers is None):
                        skip_page_zero = True
                        continue
                    else:
                        break

            transactions_prompt = self._get_prompt(prompt_type=prompt_type)
            if header_ocr_result:
                transactions_prompt = transactions_prompt.replace(
                    "#### TABLE HEADERS ####", json.dumps(header_ocr_result.content)
                )

            self.logger.info("Got Prompt")
            results = list()

            if skip_page_zero:
                image_bytes_list = image_bytes_list[
                    1:
                ]  # page zero often has account summary info
            results = self._parallel_ocr(
                image_bytes_list=image_bytes_list, prompt=transactions_prompt
            )

            # for i, image_bytes in enumerate(image_bytes_list):
            #     self.logger.info(f"Starting OCR for page: {i}")
            #     results.append(self._ocr(image_bytes=image_bytes, prompt=transactions_prompt))
            return results
        except Exception as e:
            self.logger.exception(e)
            return []

    def _get_image_block(self, image: bytes, image_content_type: ImageType) -> dict:
        return {
            "image": {
                "format": image_content_type,
                "source": {
                    "bytes": image,
                },
            }
        }

    def image_analysis(
        self, images: List[bytes], prompt: str, image_content_type: ImageType
    ) -> OCRResult:

        system_prompt = [{"text": prompt}]
        message = [
            {
                "role": "user",
                "content": [
                    self._get_image_block(
                        image=image, image_content_type=image_content_type
                    )
                    for image in images
                ],
            }
        ]
        response = self.client.converse(
            modelId=self.model_id, messages=message, system=system_prompt
        )

        result = {}
        result["content"] = Bedrock.extract_json_from_markdown(
            text = response["output"]["message"]["content"][0]["text"]
        )
        result["token_usage"] = response["usage"]
        return OCRResult.from_dict(data=result)
